import unittest
import os
import threading
from functools import partial
from hmi_simulator import run_server
from hmi_simulator.utility import func_generator
from hmi_simulator.utility.properties_constants import *
from hmi_simulator.modbus import updater as modbus_updater, server as modbus_server
from pymodbus.client.sync import ModbusTcpClient as ModbusClient


# define thread generator class for integration tests
class NewThread(object):
    def __init__(self, function_scheduled):
        self.function_scheduled = function_scheduled

        thread = threading.Thread(target=self.run, args=())
        thread.daemon = True
        thread.start()

    def run(self):
        while True:
            self.function_scheduled()


# Unit Tests
class HmiTest(unittest.TestCase):

    def test_config_file_loading(self):
        # set test configfile path
        dirname = os.path.dirname(__file__)
        configfile = os.path.join(dirname, 'resources/test_config.yaml')
        expected_properties = {
            'modbus': {'server': 'localhost', 'port': 2502, 'listen_interface': '0.0.0.0', 'slave_unit': 0,
                       'total_di_registers': 10, 'total_co_registers': 10, 'total_hr_registers': 3010,
                       'total_ir_registers': 10},
            'modbus_server_info' : {'vendor_name': 'ImolaInformatica', 'product_code': 'XYZ145', 'vendor_url': 'https://www.imolainformatica.it',
                                    'product_name': 'HMIsimulator', 'model_name': 'HMIsimulator', 'revision': '0.1'},
            'updater_jobs': {'tempi_attuatori': {'start_register': 1, 'stop_register': 29, 'frequency': 5},
                             'temperature': {'start_register': 1000, 'stop_register': 1027, 'frequency': 10},
                             'temperature_stampo': {'start_register': 2000, 'stop_register': 2071, 'frequency': 10},
                             'temperature_acqua': {'start_register': 3000, 'stop_register': 3006, 'frequency': 10}}}
        # load file
        properties = run_server.load_config(configfile)
        self.assertEqual(properties, expected_properties)

    def test_rising_saw_func_generator(self):
        # this test verifies that func generator works correctly, incrementing puntual value by 1
        amplitude = 10
        value = 1
        min_saw_value = 0
        expected_value = 2

        generated_value = func_generator.saw(amplitude, min_saw_value, value)
        # verify correct incrementation
        self.assertEqual(generated_value, expected_value)

    def test_resetting_saw_func_generator(self):
        # this test verifies that func generator correctly reset value to 0 in order to represent a saw graph
        amplitude = 10
        value = 10
        min_saw_value = 0
        expected_value = 0

        generated_value = func_generator.saw(amplitude,min_saw_value, value)
        # verify correct incrementation
        self.assertEqual(generated_value, expected_value)

    def test_modbus_server_startup_and_update(self):
        default_register_value = 3

        # set test configfile path
        dirname = os.path.dirname(__file__)
        configfile = os.path.join(dirname, 'resources/test_config.yaml')
        properties = run_server.load_config(configfile)

        # run server in separate thread
        NewThread(partial(modbus_server.run_server, properties))

        # query register and verify default value
        modbus_server_port = properties[MODBUS][PORT]
        modbus_server_addr = properties[MODBUS][SERVER]
        modbus_slave_unit = properties[MODBUS][SLAVE_UNIT]

        client = ModbusClient(modbus_server_addr, port=modbus_server_port)
        client.connect()
        register_value = client.read_holding_registers(1, 1, unit=modbus_slave_unit)
        # verify default value
        self.assertEqual(register_value.registers[0], default_register_value)

        client.close()

        # test update function
        target_job = 'tempi_attuatori'
        modbus_updater.update_holding_registers(target_job, properties)
        # verify register update
        start_register = properties[UPDATER_JOBS][target_job][START_REGISTER]
        stop_register = properties[UPDATER_JOBS][target_job][STOP_REGISTER]

        expected_register_value = 4
        client = ModbusClient(modbus_server_addr, port=modbus_server_port)
        client.connect()

        for register in range(start_register, stop_register +1):
            register_value = client.read_holding_registers(register, 1, unit=modbus_slave_unit)
            self.assertEqual(register_value.registers[0], expected_register_value)

        client.close()


if __name__ == '__main__':
    unittest.main(verbosity=3)
